#ifndef DATA_H
#define DATA_H

#include <string>
#include <map>

#define PAGESIZE 4096

using namespace std;

class AddressStats;
class PageStats;

typedef map<unsigned long, PageStats> PageMap;
typedef map<unsigned long, AddressStats> AddressMap;
typedef map<int, bool> ProcessMap;

/**
 * Helper function to page align a pointer
 */
void* pageAlign(void* p) {
   return reinterpret_cast<void*>((((int)p) & ~(PAGESIZE-1)));
}

/**
 * A random address has this type
 */
class AddressStats {
   public:
      AddressStats(void* _addr) {
         addr = _addr;
         access_freq = 1;
      }

      void* getAddr() const {return addr;}
      unsigned long getAccessFreq() const {return access_freq;}

      void accessed() {
         access_freq++;
      }

   private:
      void* addr;
      unsigned long access_freq;
};

/**
 * A page has this type
 */
class PageStats {
   public:
      PageStats(void* baseaddr, int owner) {
         base = baseaddr;
         current_owner = owner;
         fault_freq = 0;
         total_access = 0;
      }

      void updateStats(int by, void* forAddr) {
         //fprintf(stderr, "owner = %d, pid = %d, forAddr = %p\n", current_owner, by, forAddr);
         if (current_owner != by) {
            current_owner = by;
            fault_freq++;
         }
         pid_data.insert(pair<int, bool>(by, true));
         AddressMap::iterator i = addr_data.find(reinterpret_cast<unsigned long>(forAddr));

         // If the address is already in the map
         if (i != addr_data.end()) {
            fprintf(stderr, "in map\n");
            i->second.accessed();
         }
         // Else add it
         else {
            AddressStats addr(forAddr);
            addr_data.insert(pair<unsigned long, AddressStats>((unsigned long) forAddr, addr));
            pid_data.insert(pair<int, bool>(by, true));
         }
         total_access++;
         return;
      }

      int getFaultCount() const {return fault_freq;}
      int getTotalAccess() const {return total_access;}
      AddressMap getAddressStats() const {return addr_data;}
      ProcessMap getPidStats() const {return pid_data;}

   private:
      void* base;
      int current_owner;
      int fault_freq;
      int total_access;

      AddressMap addr_data;
      ProcessMap pid_data;
};

#endif
