
#ifndef DATA_PARSER_H
#define DATA_PARSER_H

#include <cstdio>
#include <cstring>
#include <map>

#include "Data.h"
#include "Textfile.h"

using namespace std;

class DataParser {
   public:
      DataParser() {
         min_addr = 0;
         max_addr = 0;
         max_faults = -1;
         min_faults = -1;
         max_accesses = 0;
         max_pid = 0;
      }

      /*
       * Parse the data from an input file path
       */
      int parseFile(string const& datafile) {

         // The file as a c-style string
         char* data = textFileRead(datafile.c_str());

         // Variables we're reading from each line of input
         int pid = 0;
         unsigned long addr = 0;

         // This probably will only work with 32-bit address
         // Good one...
         char straddr[20];

         int line = 0;

         // Pure awesome
         char* tok = strtok(data, "\n");


         // While there is another line
         while (tok != NULL) {

            // Parse actual data
            if (sscanf(tok, "%d %s", &pid, straddr) <= 0) {
               fprintf(stderr, "Error: Malformed line (%d)\n", line);
               exit(1);
            }

            // Parse hex string as unsigned long
            addr = strtoul(straddr, NULL, 16);

            if (addr < min_addr || min_addr == 0) {
               min_addr = addr;
            }
            else if (addr > max_addr || max_addr == 0) {
               max_addr = addr;
            }

            if (pid > max_pid) {
               max_pid = pid;
            }

            // Post Process Data
            void* aligned = pageAlign(reinterpret_cast<void*>(addr));

            map<unsigned long, PageStats>::iterator i =
               mem_data.find(reinterpret_cast<unsigned long>(aligned));

            // If the page is already in the map
            if (i != mem_data.end()) {
               //fprintf(stderr, "Page %p already in mem_data.\n", aligned);

               i->second.updateStats(pid, (void*) addr);

               if (max_accesses < i->second.getTotalAccess()) {
                  max_accesses = i->second.getTotalAccess();
               }


            }
            // Else add it
            else {
               fprintf(stderr, "Inserting %p into mem_data.\n", aligned);
               PageStats page(aligned, pid);
               mem_data.insert(pair<unsigned long, PageStats>((unsigned long) aligned, page));
            }

            // Advance tokenizer
            tok = strtok(NULL, "\n");
            line++;
         }

         PageMap::const_iterator pages_end = mem_data.end();
         for (PageMap::const_iterator pages_itr = mem_data.begin(); pages_itr != pages_end; ++pages_itr) {
            int faults = pages_itr->second.getFaultCount();
            if (max_faults == -1) {max_faults = faults;}
            if (min_faults == -1) {min_faults = faults;}
            if (max_faults < faults) {max_faults = faults;}
            if (min_faults > faults) {min_faults = faults;}
         }


         // Return the size of the data
         return mem_data.size();
      }


      float calcAvgAccesses() {
         int count = 0;
         int acc = 0;
         PageMap::const_iterator end = mem_data.end();
         for (PageMap::const_iterator it = mem_data.begin(); it != end; ++it) {
            acc += it->second.getTotalAccess();
            count++;
         }
         return acc / count;
      }


      /*
       * Getters/Setters
       */
      unsigned long getMinAddr() {return min_addr;}
      unsigned long getMaxAddr() {return max_addr;}
      unsigned long getMaxFaults() {return max_faults;}
      unsigned long getMinFaults() {return min_faults;}
      unsigned long getMaxAccesses() {return max_accesses;}
      int getMaxPid() {return max_pid;}
      //void setMinAddr(long addr) {min_addr = addr;}
      //void setMaxAddr(long addr) {max_addr = addr;}

      PageMap getData() {return mem_data;}

   private:

      // Map of memory address -> data/stats structure
      PageMap mem_data;

      unsigned long min_addr;
      unsigned long max_addr;
      unsigned long max_faults;
      unsigned long min_faults;
      unsigned long max_accesses;
      int max_pid;
};

#endif
