#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <vector>
#include <iostream>
#include <fstream>
#include <GL/glut.h>
#include <GL/gl.h>
#include <assert.h>
#include <time.h>

/**
 * My Final Project
 */
#include "Final.h"
#include "Data.h"
#include "DataParser.h"

#define FLT_MIN 1.1754E-38F
#define FLT_MAX 1.1754E+38F

using namespace std;

/**
 * <FLAGS>
 */

bool draw_pages = true;

bool draw_ind_addr = true;

bool draw_interference = true;

float page_space = 0.1;

/**
 * </FLAGS>
 */

// The parser
DataParser parser;

// The mape of pages
PageMap page_data;

typedef map<int, float*> PidColorMap;
PidColorMap pid_colors;

float GREEN[3] = {0, 1, 0};
float YELLOW[3] = {1, 1, 0};
float ORANGE[3] = {1, 0.6, 0};
float RED[3] = {1, 0, 0};

float DARK_GREEN[3] = {0, 0.5, 0};
float DARK_YELLOW[3] = {0.7, 0.7, 0};
float DARK_ORANGE[3] = {0.7, 0.3, 0};
float DARK_RED[3] = {0.5, 0, 0};

int GW;
int GH;

//for camera changes instead
Vector3 eye, la;
char strPop[256];

//helper functions for working with vectors
Vector3 cross(const Vector3& a, const Vector3& b) {
   Vector3 c;
   c.x = a.y*b.z - a.z*b.y;
   c.y = a.z*b.x - a.x*b.z;
   c.z = a.x*b.y - a.y*b.x;

   return(c);
}

float dot(const Vector3& a, const Vector3& b) {
   return(a.x*b.x + a.y*b.y + a.z*b.z);
}

float length(const Vector3& a) {
   return(sqrt(dot(a, a)));
}

Vector3 normalize(const Vector3& a) {
   Vector3 u;
   float mag_a = length(a);
   u.x = a.x/mag_a;
   u.y = a.y/mag_a;
   u.z = a.z/mag_a;
   return(u);
}

void PrintVec(const Vector3 v1) {
   printf("vec: %f %f %f\n", v1.x, v1.y, v1.z);
}

void renderBitmapCharacher(float x, float y, float z, void *font,char *string) {
   char *c;
   glColor3f(1, 1, 1);
   glRasterPos3f(x - 1, y, z);
   for (c=string; *c != '\0'; c++) {
      glutBitmapCharacter(font, *c);
   }
}

float* clampColor(unsigned long faults) {
   float max = parser.getMaxFaults();
   float min = parser.getMinFaults();
   float step = (float) (max - min) / 5.0;
   float step1 = step * 1;
   float step2 = step * 2;
   float step3 = step * 3;
   float step4 = step * 4;
   if (faults > min+step3) {
      return RED;
   }
   if (faults > min+step2) {
      return ORANGE;
   }
   else if (faults > min+step1) {
      return YELLOW;
   }
   else {
      return GREEN;
   }
}

float* clampDarkColor(unsigned long faults) {
  float max = parser.getMaxFaults();
  float min = parser.getMinFaults();
  float step = (float) (max - min) / 5.0;
  float step1 = step * 1;
  float step2 = step * 2;
  float step3 = step * 3;
  float step4 = step * 4;
  if (faults > min+step3) {
    return DARK_RED;
  }
  if (faults > min+step2) {
    return DARK_ORANGE;
  } 
  else if (faults > min+step1) {
    return DARK_YELLOW;
  }
  else {
    return DARK_GREEN;
  } 
}

void initInterferenceColors(int pids) {
   for (int i = 0; i < pids; i++) {
      float* c = new float[3];
      c[0] = (float) (rand() % 100) / 100.0;
      c[1] = (float) (rand() % 100) / 100.0;
      c[2] = (float) (rand() % 100) / 100.0;
      pid_colors.insert(pair<int, float* const>(i, c));
   }
   return;
}

void display() {

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glMatrixMode(GL_MODELVIEW);

   glPushMatrix();
   // Set up the camera
   gluLookAt(0, 0, 2.5,
         0, 0, 0,
         0, 1, 0);
   renderBitmapCharacher(-1.3,1.3,0,(void *)GLUT_BITMAP_HELVETICA_12,strPop);
   glPopMatrix();

   glPushMatrix();
   //set up the camera, aswd keys work but no rotation
   gluLookAt(eye.x, eye.y, eye.z, la.x, la.y, la.z, 0, 1,0);

   //rotate patch just for nice angle of viewing
   glRotatef(45, 1, 0, 0);

   glPushMatrix();
   PageMap::const_iterator end = page_data.end();
   for (PageMap::const_iterator it = page_data.begin(); it != end; ++it) {

      //fprintf(stderr, "Page %p: %d faults\n", (void*) it->first, (unsigned int) it->second.getFaultCount());

      float page_length = 1.0;

      if (draw_pages) {
         // Draw Page Line
         float scale = 10.0;
         float width = (float) it->second.getTotalAccess() / parser.calcAvgAccesses() * scale;
         glLineWidth(width);
         glBegin(GL_LINES);
         //Color depends on fault frequency
         glColor3fv(clampDarkColor(it->second.getFaultCount()));
         glVertex3f(0, 0, 0);
         glVertex3f(0, 0, page_length);
         glEnd();
         glLineWidth(1);
      }


      /* The Interference Graph
      */
      float pid_step = page_length / (parser.getMaxPid() + 1);

      // Draw Process Interference Graph
      if (draw_interference) {
         ProcessMap accessors = it->second.getPidStats();
         ProcessMap::const_iterator proc_itr_end = accessors.end();
         for (ProcessMap::const_iterator proc_itr = accessors.begin(); proc_itr != proc_itr_end; ++proc_itr) {
            glLineWidth(20);
            glBegin(GL_LINES);
            glColor3fv(pid_colors.find(proc_itr->first)->second);
            glVertex3f(0, proc_itr->first * pid_step, -0.1);
            glVertex3f(0, (proc_itr->first+1) * pid_step, -0.1);
            glEnd();
            glLineWidth(1);
         }
      }

      float page_step = page_length / 4096;

      // Returns a immutable copy of the object?
      AddressMap addr_stats = it->second.getAddressStats();
      AddressMap::const_iterator end2 = addr_stats.end();
      for (AddressMap::const_iterator it2 = addr_stats.begin(); it2 != end2; ++it2) {

         // Need this to know where to draw it, should be between 0 - 4095
         int whichAddr = ((unsigned long) it2->second.getAddr()) & 0x00000FFF;

         // Used to calculat the size of the point
         float addrFreq = (float) it2->second.getAccessFreq();

         // Get the clamped color
         float* color = clampColor(it->second.getFaultCount());

         // Omit addresses that were access only once, they clutter the screen
         if (addrFreq > 1.0) {

            float scaledAddrFreq = addrFreq / 4.0;

            if (draw_pages && draw_ind_addr) {
               // Draw a line for easy viewing
               glBegin(GL_LINES);
               glColor3fv(color);
               glVertex3f(0, 0, page_step * whichAddr);
               glVertex3f(0, scaledAddrFreq, page_step * whichAddr);
               glEnd();
            }

            if (draw_ind_addr) {
               // Draw Larger Point for Caps
               glColor3fv(color);
               glPointSize(scaledAddrFreq * 20);
               glBegin(GL_POINTS);
               glVertex3f(0, scaledAddrFreq, page_step * whichAddr);
               glEnd();
            }
         }
      }
      glTranslatef(page_space, 0, 0);
   }
   glPopMatrix();


   glPopMatrix();

   glutSwapBuffers();
   glFlush();

}

void keyboard(unsigned char key, int x, int y) {
   float delta = 0.1;
   Vector3 gaze,left;
   switch(key) {
   case 'a':
      gaze.x = la.x - eye.x;
      gaze.y = la.y - eye.y;
      gaze.z = la.z - eye.z;
      left = cross(Vector3(0, 1, 0), gaze);
      eye.x += delta*left.x;
      eye.y += delta*left.y;
      eye.z += delta*left.z;
      la.x += delta*left.x;
      la.y += delta*left.y;
      la.z += delta*left.z;
      break;
   case 'd':
      gaze.x = la.x - eye.x;
      gaze.y = la.y - eye.y;
      gaze.z = la.z - eye.z;
      left = cross(Vector3(0, 1, 0), gaze);
      eye.x -= delta*left.x;
      eye.y -= delta*left.y;
      eye.z -= delta*left.z;
      la.x -= delta*left.x;
      la.y -= delta*left.y;
      la.z -= delta*left.z;
      glutPostRedisplay();
      break;
   case 's':
      gaze.x = la.x - eye.x;
      gaze.y = la.y - eye.y;
      gaze.z = la.z - eye.z;
      eye.x -= delta*gaze.x;
      eye.y -= delta*gaze.y;
      eye.z -= delta*gaze.z;
      la.x -= delta*gaze.x;
      la.y -= delta*gaze.y;
      la.z -= delta*gaze.z;
      glutPostRedisplay();
      break;
   case 'w':
      gaze.x = la.x - eye.x;
      gaze.y = la.y - eye.y;
      gaze.z = la.z - eye.z;
      eye.x += delta*gaze.x;
      eye.y += delta*gaze.y;
      eye.z += delta*gaze.z;
      la.x += delta*gaze.x;
      la.y += delta*gaze.y;
      la.z += delta*gaze.z;
      glutPostRedisplay();
      break;
   case 'p':
      draw_pages = draw_pages ? false : true;
      break;
   case 'i':
      draw_ind_addr = draw_ind_addr ? false : true;
      break;
   case 'o':
      draw_interference = draw_interference ? false : true;
      break;
   case '+':
      page_space += 0.01;
      break;
   case '-':
      page_space = page_space > 0 ? page_space - 0.01 : page_space;
      break;
   case 'q':
      exit( EXIT_SUCCESS );
      break;
   default:
      break;
   }
   glutPostRedisplay();
}

void reshape(int w, int h) {
   GW = w;
   GH = h;

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   gluPerspective(60.0, (float)w/h, 0.1, 50);
   glMatrixMode(GL_MODELVIEW);
   glViewport(0, 0, w, h);

   glutPostRedisplay();
}

int main( int argc, char** argv ) {

   glutInit(&argc, argv);

   glutInitWindowPosition(100, 100);
   glutInitWindowSize(1200, 600);
   glutCreateWindow("Memory Access/Fault Visualization");
   glClearColor(0.2, 0.2, 0.2, 0);

   glutDisplayFunc( display );
   glutReshapeFunc( reshape );
   glutKeyboardFunc(keyboard);
   glEnable(GL_DEPTH_TEST);

   GW=GH = 500;

   srand(time(NULL));

   eye.x = 0.0; eye.y = 0.0; eye.z = 4.0;
   la.x = 0; la.y = 0.0; la.z = 0;


   parser.parseFile("data/data.txt");

   // Set global variable (main sci data)
   page_data = parser.getData();

   // We need a distinct color for each process
   initInterferenceColors(parser.getMaxPid() + 1);


   // <Diagnostics>
   PageMap::const_iterator end = page_data.end();
   for (PageMap::const_iterator it = page_data.begin(); it != end; ++it) {
      ProcessMap accessors = it->second.getPidStats();
      fprintf(stderr, "Page %p: %d faults %d access by (%d) PIDs: ", (void*) it->first, it->second.getFaultCount(), it->second.getTotalAccess(), accessors.size());
      ProcessMap::const_iterator end2 = accessors.end();
      for (ProcessMap::const_iterator it2 = accessors.begin(); it2 != end2; ++it2) {
         fprintf(stderr, "%d ", it2->first);
      }
      fprintf(stderr, "\n");
   }

   sprintf(strPop, "Min/Max Addr = %p/%p\nMin/Max Faults = %d/%d\n",
         (void*) parser.getMinAddr(),
         (void*) parser.getMaxAddr(),
         parser.getMinFaults(),
         parser.getMaxFaults());

   fprintf(stderr, "Min/Max Addr = %p/%p\n", (void*)parser.getMinAddr(), (void*)parser.getMaxAddr());
   fprintf(stderr, "Min/Max Faults = %d/%d\n", parser.getMinFaults(), parser.getMaxFaults());
   fprintf(stderr, "Max PID = %d\n", parser.getMaxPid());

   // Prints assigned PID colors
   //
   //map<int, float*>::const_iterator pidcolor_end = pid_colors.end();
   //for (map<int, float*>::const_iterator pidcolor_itr = pid_colors.begin(); pidcolor_itr != pidcolor_end; ++pidcolor_itr) {
   //  float* t = pidcolor_itr->second;
   //  fprintf(stderr, "%d: <%f, %f, %f>\n", pidcolor_itr->first, t[0], t[1], t[2]);
   //}

   // </Diagnostics>


   glutMainLoop();
}


